@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtBluetooth module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2022 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qbluetoothdevicediscoveryagent_p.h"
 #include "qbluetoothdevicediscoveryagent.h"
@@ -60,6 +24,7 @@
 #include "qbluetoothuuid.h"
 
 #include <QtCore/qloggingcategory.h>
+#include <QtCore/qcoreapplication.h>
 #include <QtCore/qvector.h>
 #include <QtCore/qglobal.h>
 #include <QtCore/qstring.h>
@@ -83,18 +48,13 @@ void registerQDeviceDiscoveryMetaType()
         initDone = true;
     }
 }
-#ifdef Q_OS_MACOS
-using InquiryObjC = QT_MANGLE_NAMESPACE(DarwinBTClassicDeviceInquiry);
-#endif // Q_OS_MACOS
-
-using LEInquiryObjC = QT_MANGLE_NAMESPACE(DarwinBTLEDeviceInquiry);
 
 } //namespace
 
 QBluetoothDeviceDiscoveryAgentPrivate::QBluetoothDeviceDiscoveryAgentPrivate(const QBluetoothAddress &adapter,
                                                                              QBluetoothDeviceDiscoveryAgent *q) :
-    agentState(NonActive),
     adapterAddress(adapter),
+    agentState(NonActive),
     lowEnergySearchTimeout(DarwinBluetooth::defaultLEScanTimeoutMS),
 #ifdef Q_OS_MACOS
     requestedMethods(QBluetoothDeviceDiscoveryAgent::ClassicMethod | QBluetoothDeviceDiscoveryAgent::LowEnergyMethod),
@@ -106,15 +66,6 @@ QBluetoothDeviceDiscoveryAgentPrivate::QBluetoothDeviceDiscoveryAgentPrivate(con
     registerQDeviceDiscoveryMetaType();
 
     Q_ASSERT_X(q != nullptr, Q_FUNC_INFO, "invalid q_ptr (null)");
-
-#ifdef Q_OS_MACOS
-    IOBluetoothHostController *hostController = [IOBluetoothHostController defaultController];
-    if (!hostController || [hostController powerState] != kBluetoothHCIPowerStateON) {
-        qCCritical(QT_BT_DARWIN) << "no default host controller or adapter is off";
-        return;
-    }
-    controller.reset(hostController, DarwinBluetooth::RetainPolicy::doInitialRetain);
-#endif
 }
 
 QBluetoothDeviceDiscoveryAgentPrivate::~QBluetoothDeviceDiscoveryAgentPrivate()
@@ -123,7 +74,7 @@ QBluetoothDeviceDiscoveryAgentPrivate::~QBluetoothDeviceDiscoveryAgentPrivate()
         // We want the LE scan to stop as soon as possible.
         if (dispatch_queue_t leQueue = DarwinBluetooth::qt_LE_queue()) {
             // Local variable to be retained ...
-            LEInquiryObjC *inq = inquiryLE.getAs<LEInquiryObjC>();
+            DarwinBTLEDeviceInquiry *inq = inquiryLE.getAs<DarwinBTLEDeviceInquiry>();
             dispatch_sync(leQueue, ^{
                 [inq stop];
             });
@@ -147,15 +98,34 @@ void QBluetoothDeviceDiscoveryAgentPrivate::start(QBluetoothDeviceDiscoveryAgent
     using namespace DarwinBluetooth;
 
     Q_ASSERT(!isActive());
-    Q_ASSERT(lastError != QBluetoothDeviceDiscoveryAgent::InvalidBluetoothAdapterError);
     Q_ASSERT(methods & (QBluetoothDeviceDiscoveryAgent::ClassicMethod
                         | QBluetoothDeviceDiscoveryAgent::LowEnergyMethod));
 
 #ifdef Q_OS_MACOS
     if (!controller) {
-        setError(QBluetoothDeviceDiscoveryAgent::InvalidBluetoothAdapterError);
-        emit q_ptr->errorOccurred(lastError);
-        return;
+        IOBluetoothHostController *hostController = [IOBluetoothHostController defaultController];
+        if (!hostController) {
+            qCWarning(QT_BT_DARWIN) << "No default Bluetooth controller found";
+            setError(QBluetoothDeviceDiscoveryAgent::InvalidBluetoothAdapterError);
+            emit q_ptr->errorOccurred(lastError);
+            return;
+        } else if ([hostController powerState] != kBluetoothHCIPowerStateON) {
+            qCWarning(QT_BT_DARWIN) << "Default Bluetooth controller is OFF";
+            setError(QBluetoothDeviceDiscoveryAgent::PoweredOffError);
+            emit q_ptr->errorOccurred(lastError);
+            return;
+        } else if (!adapterAddress.isNull()) {
+            // If user has provided the local address, check that it matches with the actual
+            NSString *const hciAddress = [hostController addressAsString];
+            if (adapterAddress != QBluetoothAddress(QString::fromNSString(hciAddress))) {
+                qCWarning(QT_BT_DARWIN) << "Provided address" << adapterAddress
+                                        << "does not match with adapter:" << hciAddress;
+                setError(QBluetoothDeviceDiscoveryAgent::InvalidBluetoothAdapterError);
+                emit q_ptr->errorOccurred(lastError);
+                return;
+            }
+        }
+        controller.reset(hostController, DarwinBluetooth::RetainPolicy::doInitialRetain);
     }
 #endif // Q_OS_MACOS
 
@@ -173,7 +143,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::start(QBluetoothDeviceDiscoveryAgent
         qCWarning(QT_BT_DARWIN)
                 << "A proper Info.plist with NSBluetoothAlwaysUsageDescription "
                    "entry is required, cannot start device discovery";
-        setError(QBluetoothDeviceDiscoveryAgent::UnsupportedDiscoveryMethod);
+        setError(QBluetoothDeviceDiscoveryAgent::MissingPermissionsError);
         emit q_ptr->errorOccurred(lastError);
         return;
     }
@@ -212,7 +182,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startClassic()
 
     if (!inquiry) {
         // The first Classic scan for this DDA.
-        inquiry.reset([[InquiryObjC alloc] initWithDelegate:this],
+        inquiry.reset([[DarwinBTClassicDeviceInquiry alloc] initWithDelegate:this],
                       DarwinBluetooth::RetainPolicy::noInitialRetain);
 
         if (!inquiry) {
@@ -226,7 +196,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startClassic()
 
     agentState = ClassicScan;
 
-    const IOReturn res = [inquiry.getAs<InquiryObjC>() start];
+    const IOReturn res = [inquiry.getAs<DarwinBTClassicDeviceInquiry>() start];
     if (res != kIOReturnSuccess) {
         setError(res, QCoreApplication::translate(DEV_DISCOVERY, DD_NOT_STARTED));
         agentState = NonActive;
@@ -257,7 +227,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startLE()
                       this, DeviceMemFunPtr(&QBluetoothDeviceDiscoveryAgentPrivate::deviceFound));
 
     // Check queue and create scanner:
-    inquiryLE.reset([[LEInquiryObjC alloc] initWithNotifier:notifier.get()],
+    inquiryLE.reset([[DarwinBTLEDeviceInquiry alloc] initWithNotifier:notifier.get()],
                     DarwinBluetooth::RetainPolicy::noInitialRetain);
     if (inquiryLE)
         notifier.release(); // Whatever happens next, inquiryLE is already the owner ...
@@ -274,7 +244,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startLE()
     // Now start in on LE queue:
     agentState = LEScan;
     // We need the local variable so that it's retained ...
-    LEInquiryObjC *inq = inquiryLE.getAs<LEInquiryObjC>();
+    DarwinBTLEDeviceInquiry *inq = inquiryLE.getAs<DarwinBTLEDeviceInquiry>();
     dispatch_async(leQueue, ^{
         [inq startWithTimeout:lowEnergySearchTimeout];
     });
@@ -296,7 +266,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::stop()
 
 #ifdef Q_OS_MACOS
     if (agentState == ClassicScan) {
-        const IOReturn res = [inquiry.getAs<InquiryObjC>() stop];
+        const IOReturn res = [inquiry.getAs<DarwinBTClassicDeviceInquiry>() stop];
         if (res != kIOReturnSuccess) {
             qCWarning(QT_BT_DARWIN) << "failed to stop";
             startPending = prevStart;
@@ -312,7 +282,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::stop()
         dispatch_queue_t leQueue(qt_LE_queue());
         Q_ASSERT(leQueue);
         // We need the local variable so that it's retained ...
-        LEInquiryObjC *inq = inquiryLE.getAs<LEInquiryObjC>();
+        DarwinBTLEDeviceInquiry *inq = inquiryLE.getAs<DarwinBTLEDeviceInquiry>();
         dispatch_sync(leQueue, ^{
             [inq stop];
         });
@@ -409,7 +379,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::setError(QBluetoothDeviceDiscoveryAg
 {
     lastError = error;
 
-    if (text.length() > 0) {
+    if (!text.isEmpty()) {
         errorString = text;
     } else {
         switch (lastError) {
@@ -428,6 +398,9 @@ void QBluetoothDeviceDiscoveryAgentPrivate::setError(QBluetoothDeviceDiscoveryAg
         case QBluetoothDeviceDiscoveryAgent::UnsupportedPlatformError:
             errorString = QCoreApplication::translate(DEV_DISCOVERY, DD_NOTSUPPORTED);
             break;
+        case QBluetoothDeviceDiscoveryAgent::MissingPermissionsError:
+            errorString = QCoreApplication::translate(DEV_DISCOVERY, DD_MISSING_PERMISSION);
+            break;
         case QBluetoothDeviceDiscoveryAgent::UnknownError:
         default:
             errorString = QCoreApplication::translate(DEV_DISCOVERY, DD_UNKNOWN_ERROR);
@@ -438,7 +411,8 @@ void QBluetoothDeviceDiscoveryAgentPrivate::setError(QBluetoothDeviceDiscoveryAg
 void QBluetoothDeviceDiscoveryAgentPrivate::LEinquiryError(QBluetoothDeviceDiscoveryAgent::Error error)
 {
     Q_ASSERT(error == QBluetoothDeviceDiscoveryAgent::PoweredOffError
-             || error == QBluetoothDeviceDiscoveryAgent::UnsupportedDiscoveryMethod);
+             || error == QBluetoothDeviceDiscoveryAgent::UnsupportedDiscoveryMethod
+             || error == QBluetoothDeviceDiscoveryAgent::MissingPermissionsError);
 
     inquiryLE.reset();
 
@@ -502,7 +476,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::deviceFound(const QBluetoothDeviceIn
 #else
             true;
 #endif // Q_OS_MACOS
-    for (int i = 0, e = discoveredDevices.size(); i < e; ++i) {
+    for (qsizetype i = 0, e = discoveredDevices.size(); i < e; ++i) {
         if (isLE) {
             if (discoveredDevices[i].deviceUuid() == newDeviceInfo.deviceUuid()) {
                 QBluetoothDeviceInfo::Fields updatedFields = QBluetoothDeviceInfo::Field::None;
@@ -521,6 +495,14 @@ void QBluetoothDeviceDiscoveryAgentPrivate::deviceFound(const QBluetoothDeviceIn
                     updatedFields.setFlag(QBluetoothDeviceInfo::Field::ManufacturerData);
                 }
 
+                if (discoveredDevices[i].serviceData() != newDeviceInfo.serviceData()) {
+                    qCDebug(QT_BT_DARWIN) << "Updating service data for" << newDeviceInfo.address();
+                    const QList<QBluetoothUuid> keys = newDeviceInfo.serviceIds();
+                    for (auto key : keys)
+                        discoveredDevices[i].setServiceData(key, newDeviceInfo.serviceData(key));
+                    updatedFields.setFlag(QBluetoothDeviceInfo::Field::ServiceData);
+                }
+
                 if (lowEnergySearchTimeout > 0) {
                     if (discoveredDevices[i] != newDeviceInfo) {
                         discoveredDevices.replace(i, newDeviceInfo);
